/* File: main-xxx.c */

/* Purpose: Sample visual module for Angband 2.7.9 */

/* Most of the code in this file (all not appearing in main-xxx.c) is
 * Copyright 2001 Gregory Velichansky (hmaon@bumba.net)
 * You may use it under the terms of the standard Angband license or
 * the GNU General Public License Version 2 or greater.
 * Ben Harrison's copyright appears in the file COPYING and at last
 * check allows you to use his code under the GPL as well.
 *
 * The Angband license states:
 * This software may be copied and distributed for educational, research,
 * and not for profit purposes provided that this copyright and statement
 * are included in all such copies.  Other copyrights may also apply.
 * 
 * 
	main-sdl.c - SDL (http://libsdl.org) Term implementation for Angband.
	Copyright (C) 2001  Gregory Velichansky (hmaon@bumba.net)
	Portions Copyright (C) 1997 Ben Harrison
	(see the file COPYING in the latest distribution of the Angband source code)

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, please see 
	http://www.gnu.org/copyleft/gpl.html

*/



/*
 * This file originally written by "Ben Harrison (benh@voicenet.com)".
 *
 */


#include "angband.h"


#ifdef USE_SDL 

#include "SDL.h"
#include <string.h>

#include <math.h> /* for scaling blits */

/* perhaps this should be in config.h:  (not if it's distributed on its own) */
#define USE_HEX_FONTS
#define SDL_MAGIC_KEYS
#define SDL_ALLOW_FLOAT

/* META_ADDRESS is probably only defined in MAngband, right? XXX XXX XXX */
#if defined(MANGBAND) || defined(META_ADDRESS) 
#define SDL_CONSOLE_HACK
#define OLD_TERM_28X
#endif

#ifdef USE_GRAPHICS
#define USE_TILE_GRAPHICS
#endif


#ifdef SDL_CONSOLE_HACK
/* The MAngband console will not link without this because c-util.o 
 * cannot be linked in. Perhaps it belongs in z-util.c?
 */
errr path_build_sdl(char *buf, int max, cptr path, cptr file)
{
        /* Special file */
        if (file[0] == '~')
        {
                /* Use the file itself */
                strnfmt(buf, max, "%s", file);
        }

        /* Absolute file, on "normal" systems */
        else if (prefix(file, PATH_SEP) && !streq(PATH_SEP, ""))
        {
                /* Use the file itself */
                strnfmt(buf, max, "%s", file);
        }

        /* No path given */
        else if (!path[0])
        {
                /* Use the file itself */
                strnfmt(buf, max, "%s", file);
        }

        /* Path and File */
        else
        {
                /* Build the new path */
                strnfmt(buf, max, "%s%s%s", path, PATH_SEP, file);
        }

        /* Success */
        return (0);
}

void init_file_paths_sdl(char *path)
{
        char *tail;


        /*** Free everything ***/

        /* Free the main path */
        string_free(ANGBAND_DIR);

        /* Free the sub-paths */
        string_free(ANGBAND_DIR_APEX);
        string_free(ANGBAND_DIR_BONE);
        string_free(ANGBAND_DIR_DATA);
        string_free(ANGBAND_DIR_EDIT);
        string_free(ANGBAND_DIR_FILE);
        string_free(ANGBAND_DIR_HELP);
        string_free(ANGBAND_DIR_INFO);
        string_free(ANGBAND_DIR_SAVE);
        string_free(ANGBAND_DIR_USER);
        string_free(ANGBAND_DIR_XTRA);


        /*** Prepare the "path" ***/

        /* Hack -- save the main directory */
        ANGBAND_DIR = string_make(path);

        /* Prepare to append to the Base Path */
        tail = path + strlen(path);


#ifdef VM


        /*** Use "flat" paths with VM/ESA ***/

        /* Use "blank" path names */
        ANGBAND_DIR_APEX = string_make("");
        ANGBAND_DIR_BONE = string_make("");
        ANGBAND_DIR_DATA = string_make("");
        ANGBAND_DIR_EDIT = string_make("");
        ANGBAND_DIR_FILE = string_make("");
        ANGBAND_DIR_HELP = string_make("");
        ANGBAND_DIR_INFO = string_make("");
        ANGBAND_DIR_SAVE = string_make("");
        ANGBAND_DIR_USER = string_make("");
        ANGBAND_DIR_XTRA = string_make("");


#else /* VM */


        /*** Build the sub-directory names ***/

        /* Build a path name */
        strcpy(tail, "apex");
        ANGBAND_DIR_APEX = string_make(path);

        /* Build a path name */
        strcpy(tail, "bone");
        ANGBAND_DIR_BONE = string_make(path);

        /* Build a path name */
        strcpy(tail, "data");
        ANGBAND_DIR_DATA = string_make(path);

        /* Build a path name */
        strcpy(tail, "edit");
        ANGBAND_DIR_EDIT = string_make(path);

        /* Build a path name */
        strcpy(tail, "file");
        ANGBAND_DIR_FILE = string_make(path);

        /* Build a path name */
        strcpy(tail, "help");
        ANGBAND_DIR_HELP = string_make(path);

        /* Build a path name */
        strcpy(tail, "info");
        ANGBAND_DIR_INFO = string_make(path);

        /* Build a path name */
        strcpy(tail, "save");
        ANGBAND_DIR_SAVE = string_make(path);

        /* Build a path name */
        strcpy(tail, "user");
        ANGBAND_DIR_USER = string_make(path);

        /* Build a path name */
        strcpy(tail, "xtra");
        ANGBAND_DIR_XTRA = string_make(path);

#endif /* VM */


#ifdef NeXT

        /* Allow "fat binary" usage with NeXT */
        if (TRUE)
        {
                cptr next = NULL;

# if defined(m68k)
                next = "m68k";
# endif

# if defined(i386)
                next = "i386";
# endif

# if defined(sparc)
                next = "sparc";
# endif

# if defined(hppa)
                next = "hppa";
# endif

                /* Use special directory */
                if (next)
                {
                        /* Forget the old path name */
                        string_free(ANGBAND_DIR_DATA);

                        /* Build a new path name */
                        sprintf(tail, "data-%s", next);
                        ANGBAND_DIR_DATA = string_make(path);
                }
        }

#endif /* NeXT */

}

#define path_build path_build_sdl
#define init_file_paths init_file_paths_sdl
#endif


/*
 * Extra data to associate with each "window"
 *
 * Each "window" is represented by a "term_data" structure, which
 * contains a "term" structure, which contains a pointer (t->data)
 * back to the term_data structure.
 */

typedef struct font_data font_data; /* must be here to avoid fwd. ref. */


/* tile graphics are very very similar to bitmap fonts. */
typedef struct font_data graf_tiles; 


typedef struct term_data term_data;

struct term_data
{
	term		t;

	cptr		name;

	int width, height, bpp; 
	/* XXX width, height, and bpp are only used to provide hints to 
	 * Term_init_sdl(). Consider them write-only values!
	 * If you need  to read the actual window dimensions, use
	 * face->w, face->h, and face->format
	 */

	SDL_Surface *face;

	font_data *fd;

	graf_tiles *gt;

	Uint8 w, h; /* width and height of an individual 2D element */

	Uint32 flags;
};



/*
 * One "term_data" for each "window"
 *
 * XXX XXX XXX The only "window" which you MUST support is the
 * main "term_screen" window, the rest are optional.  If you only
 * support a single window, then most of the junk involving the
 * "term_data" structures is actually not needed, since you can
 * use global variables.  But you should avoid global variables
 * when possible as a general rule...
 */
static term_data screen;
static term_data mirror;
static term_data recall;
static term_data choice;


#if 0	/* Fix the syntax below */

/*
 * XXX XXX XXX The "color" array for the visual module
 *
 * This table should be used in whetever way is necessary to
 * convert the Angband Color Indexes into the proper "color data"
 * for the visual system.  On the Macintosh, these are arrays of
 * three shorts, on the IBM, these are combinations of the eight
 * basic color codes with optional "bright" bits, on X11, these
 * are actual "pixel" codes extracted from another table which
 * contains textual color names.
 *
 * The Angband Color Set (0 to 15):
 *   Black, White, Slate, Orange,    Red, Blue, Green, Umber
 *   D-Gray, L-Gray, Violet, Yellow, L-Red, L-Blue, L-Green, L-Umber
 *
 * Colors 8 to 15 are basically "enhanced" versions of Colors 0 to 7.
 *
 * As decribed in one of the header files, in a perfect world, the
 * colors below should fit a nice clean "quartered" specification
 * in RGB codes, but this must often be Gamma Corrected.  The 1/4
 * parts of each Red,Green,Blue are shown in the comments below,
 * again, these values are *before* gamma correction.
 */
static local_color_data_type color_data[16] =
{
	/* XXX XXX XXX 0,0,0 */,		/* TERM_DARK */
	/* XXX XXX XXX 4,4,4 */,		/* TERM_WHITE */
	/* XXX XXX XXX 2,2,2 */,		/* TERM_SLATE */
	/* XXX XXX XXX 4,2,0 */,		/* TERM_ORANGE */
	/* XXX XXX XXX 3,0,0 */,		/* TERM_RED */
	/* XXX XXX XXX 0,2,1 */,		/* TERM_GREEN */
	/* XXX XXX XXX 0,0,4 */,		/* TERM_BLUE */
	/* XXX XXX XXX 2,1,0 */,		/* TERM_UMBER */
	/* XXX XXX XXX 1,1,1 */,		/* TERM_L_DARK */
	/* XXX XXX XXX 3,3,3 */,		/* TERM_L_WHITE */
	/* XXX XXX XXX 4,0,4 */,		/* TERM_VIOLET */
	/* XXX XXX XXX 4,4,0 */,		/* TERM_YELLOW */
	/* XXX XXX XXX 4,0,0 */,		/* TERM_L_RED */
	/* XXX XXX XXX 0,4,0 */,		/* TERM_L_GREEN */
	/* XXX XXX XXX 0,4,4 */,		/* TERM_L_BLUE */
	/* XXX XXX XXX 3,2,1 */		/* TERM_L_UMBER */
};

#endif



/*** Special functions ***/


#ifdef SDL_ALLOW_FLOAT
/* The most pedantic-a%& getpixel and putpixel ever, hopefully. */
errr SDL_GetPixel (SDL_Surface *f, Uint32 x, Uint32 y, Uint8 *r, Uint8 *g, Uint8 *b)
{
	/*const Uint32 mask[] = {0x0, 0xff, 0xffff, 0xffffff, 0xffffffff};*/
	Uint32 pixel;

	Uint8 *pp;

	int n; /* general purpose 'n'. */

	if (f == NULL) return -1;

	pp = (Uint8 *) f->pixels;

	if (x >= f->w || y >= f->h) return -1;

	pp += (f->pitch * y);

	pp += (x * f->format->BytesPerPixel);

	/* we do not lock the surface here, it would be inefficient XXX */
	/* this reads the pixel as though it was a big-endian integer XXX */
	/* I'm trying to avoid reading part the end of the pixel data by
	 * using a data-type that's larger than the pixels */
	for (n = 0, pixel = 0; n < f->format->BytesPerPixel; ++n, ++pp)
	{
		pixel <<= 8;
		pixel |= *pp;
	}

	SDL_GetRGB(pixel, f->format, r, g, b);
	return 0;
}

/* This function looks remarkably similar to the one above. Yes, it's cut
 * and paste. */
errr SDL_PutPixel (SDL_Surface *f, Uint32 x, Uint32 y, Uint8 r, Uint8 g, Uint8 b)
{
	Uint32 pixel;

	Uint8 *pp;

	int n; 

	if (f == NULL) return -1;

	pp = (Uint8 *) f->pixels;

	if (x >= f->w || y >= f->h) return -1;

	pp += (f->pitch * y);

	pp += (x * f->format->BytesPerPixel);

	pixel = SDL_MapRGB(f->format, r, g, b);

	for (n = 0; n < f->format->BytesPerPixel; ++n, ++pp)
	{
		*pp = (Uint8) (pixel & 0xFF);
		pixel >>= 8;
	}

	return 0;
}

/* The following is a function to perform a Blit while magnifying */
/* Anti-aliasing is performed. :) */
/* It is probably very SLOW on most systems. Use it for pre-processing. XXX */
/* a Blit while shrinking is handled by a different function */
errr SDL_StretchBlit(SDL_Surface *src, SDL_Rect *sr, SDL_Surface *dst, SDL_Rect *dr)
{
	double sx, sy; /* current source x and y */
	Uint32 isx, isy; /* temp. values for convenience in calculation code */
	double dsx, dsy; /* source increment, per increment of 1 in destination */
	double wx, wy; /* temp. weight values for the color mixing calculations */
	double weight; /* temp. weight of pixel */

	/* coordinates to get pixels from: ... */
#define gsx (isx >= lx ? isx-1 : isx)
#define gsy (isy >= ly ? isy-1 : isy)

	Uint32 tx, ty; /* "to" x and y. "dx, dy" would be too confusing. */
	Uint32 lx, ly; /* end x and y in destination, not inclusive */

	double r, g, b; /* temporary values on which we perform calculations */
	Uint8 ir, ig, ib; /* same here. */

	double scale = hypot(1, 1); /* the scale of the weight... */

	if (src == NULL || sr == NULL || dst == NULL || dr == NULL) return -1;

	/* these are meaningless and would cause a divide by zero: */
	if (!dr->w || !dr->h) return -1; 

	dsx = ((double) sr->w) / dr->w;
	dsy = ((double) sr->h) / dr->h;
	printf("dsx, dsy %f, %f\n", (double) dsx, (double) dsy);

	lx = dr->x + dr->w - 1; /* FIXME */
	ly = dr->y + dr->h - 1;

	for (ty = dr->y, sy = (double)sr->y; ty < ly; ++ty, sy+=dsy)
	{
		for (tx = dr->x, sx = (double)sr->x; tx < lx; ++tx, sx+=dsx)
		{
			/* here we must consider four pixels and mix them together */
			/* the further away we are from a pixel, the less weight it has
			 * when we mix in its color. Hence the "1 - hypot(..." etc.
			 */
			/* 
			 * upper left pixel 
			 */
			wx = (sx - floor(sx)); 
			wy = (sy - floor(sy));

			isx = (Uint32) floor(sx);
			isy = (Uint32) floor(sy);

			if (SDL_GetPixel(src, gsx, gsy, &ir, &ig, &ib)) return -1;
			
			weight = 1 - hypot(wx, wy) / scale; /* Replace with a lookup table! XXX FIXME */
			r = weight * ir;
			g = weight * ig;
			b = weight * ib;

			/* 
			 * upper right pixel 
			 */
			wx = 1 - wx;
			isx += 1;

			if (SDL_GetPixel(src, gsx, gsy, &ir, &ig, &ib)) return -1;
			
			weight = 1 - hypot(wx, wy) / scale; /* Replace with a lookup table! XXX FIXME */
			r += weight * ir;
			g += weight * ig;
			b += weight * ib;
	
			/* 
			 * lower right pixel 
			 */
			wy = 1 - wy;
			isy += 1;

			if (SDL_GetPixel(src, gsx, gsy, &ir, &ig, &ib)) return -1;
			
			weight = 1 - hypot(wx, wy) / scale; /* Replace with a lookup table! XXX FIXME */
			r += weight * ir;
			g += weight * ig;
			b += weight * ib;

			/*
			 * lower left pixel
			 */
			wx = 1 - wx;
			isx -= 1;

			if (SDL_GetPixel(src, gsx, gsy, &ir, &ig, &ib)) return -1;
			
			weight = 1 - hypot(wx, wy) / scale; /* Replace with a lookup table! XXX FIXME */
			r += weight * ir;
			g += weight * ig;
			b += weight * ib;
	
			if (r > 255.0 || g > 255.0 || b > 255.0) {plog("mixing error!");
			printf("Values: %f, %f, %f\n", (double)r, (double)g, (double)b);
			*((char *)0) = 0;}
			if (r > 255.0) r = 255.0;
			if (g > 255.0) g = 255.0;
			if (b > 255.0) b = 255.0;
			ir = (Uint8) r;
			ig = (Uint8) g;
			ib = (Uint8) b;

			SDL_PutPixel(dst, tx, ty, ir, ig, ib);
		}
	}
}



#endif /* SDL_ALLOW_FLOAT */

static SDL_Color color_data_sdl[16] =
{
	{0, 0, 0, 0}, 
	{4, 4, 4, 0}, 
	{2, 2, 2, 0}, 
	{4, 2, 0, 0}, 
	{3, 0, 0, 0}, 
	{0, 2, 1, 0}, 
	{0, 0, 4, 0}, 
	{2, 1, 0, 0}, 
	{1, 1, 1, 0}, 
	{3, 3, 3, 0}, 
	{4, 0, 4, 0}, 
	{4, 4, 0, 0}, 
	{4, 0, 0, 0}, 
	{0, 4, 0, 0}, 
	{0, 4, 4, 0}, 
	{3, 2, 1, 0}
};

/*#define SCALETOCOLOR(x) (x=((x)*63+((x)-1)))*/
#define ScaleToColor(x) ((x)=((x)*60)+15)
void init_color_data_sdl() {
	Uint8 i;

	for (i = 0; i < 16; ++i) {
		color_data_sdl[i].unused = 255; /* no reason. */
		if(!color_data_sdl[i].r && !color_data_sdl[i].g && !color_data_sdl[i].b)
			continue;
		ScaleToColor(color_data_sdl[i].r);
		ScaleToColor(color_data_sdl[i].g);
		ScaleToColor(color_data_sdl[i].b);
	}
}


/*
 * XXX XXX XXX You may need an event handler here
 */

/* 
 * XXX XXX XXX Nope. 
 */




/*
 *
 * Text drawing code.
 *
 */


struct font_data 
{
	/*
	 * to find a character:
	 * x = character * w
	 */
	SDL_Surface *face;
	/* 
	 * font metrics.
	 * Obviously, the font system is very minimalist.
	 */
	Uint8 w;
	Uint8 h;

	Uint8 dw; /* width and height of font on destination surface */
	Uint8 dh;
};

font_data screen_font;
graf_tiles screen_tiles;


/* The following function will extract height and width info from a filename
 * such as 16x16.xyz or 8X13.bar or even argle8ook16.foo
 */

errr strmetrics(const char *str, Uint32 *w, Uint32 *h)
{
	char buf[1024];
	char *s = buf;
	char *tok;
	char *numeric = "0123456789";

	size_t l; /* length of numeric string */

	if (!str || !w || !h) return -1;

	if (strlen(str) < 3) return -1; /* must have room for at least "1x1" */

	strncpy(buf, str, 1023);
	buf[1023] = '\0';

	tok = strpbrk(buf, numeric);
	if (!tok) return -1;

	l = strspn(tok, numeric);
	if (!l) return -1;

	tok[l] = '\0';

	s = tok + l + 1;

	if(!sscanf(tok, "%d", w)) return -1;

	/* next token */
	tok = strpbrk(s, numeric);
	if (!tok) return -1;

	l = strspn(tok, numeric);
	if (!l) return -1;

	tok[l] = '\0';
	/* no need to set s since this is the last token */

	if(!sscanf(tok, "%d", h)) return -1;

	return 0;

}


#ifdef USE_HEX_FONTS
/*
 * Load a HEX font.
 * See http://czyborra.com/unifont/
 *
 * XXX Note. Your file should not be all full-width glyphs. At least one
 * half-width glyph must be present for my lame algorithm to work.
 * It is OK to have all half-width glyphs.
 *
 * FIXME
 * BUGS: There is no attempt made at figuring out a righteous bounding box.
 *	      Certain HEX fonts can be *wider* than 16 pixels. They don't work.
 *
 *	What we need is a BDF loader. It's not a high priority though.
 *
 */

#define highhextoi(x) (strchr("ABCDEF", (x))? 0xA + ((x)-'A'):0)
#define hexchartoi(x) (strchr("0123456789", (x))? (x)-'0' : highhextoi((x)))

#ifndef MAX_HEX_FONT_LINE
#define MAX_HEX_FONT_LINE 1024
#endif

errr load_HEX_font_sdl(font_data *fd, cptr filename, bool justmetrics)
{
	FILE *f;

	char buf[1036]; /* 12 (or 11? ;->)extra bytes for good luck. */

	char gs[MAX_HEX_FONT_LINE]; /* glyph string */

	Uint32 i,j;

	errr fail = 0; /* did we fail? */

	Uint32 x; /* current x in fd->face */
	Uint32 y; /* current y in fd->face */

	Uint32 gx; /* current glyph x */

	Uint32 n; /* current nibble or byte or whatever data from file */

	Uint32 pos; /* position in the nasty string */

	Uint32 bytesdone; /* bytes processed */

	Uint32 mw, mh; /* for strmetrics() */



	/* check font_data */
	if (fd->w || fd->h || fd->face) return 1; /* dealloc it first, dummy. */

	/* Build the filename */
	path_build(buf, 1024, ANGBAND_DIR_XTRA, filename);

	f = fopen(buf, "r");

	if (!f) 
	{
		plog(format("Couldn't open: %s", buf));
		return -1;
	}

	

	/* try hard to figure out the font metrics */

	while (fgets(gs, MAX_HEX_FONT_LINE, f) != NULL)
	{
		i = strlen(gs);

		if (gs[i-1] == '\n') i--;
		if (gs[i-1] == '\r') i--;
		
		i -= 5; /* each line begins with 1234: */

		/* now i is the number of nibbles in the line */

		if (i & 1)
		{
			plog("Error in HEX line measurment. Report to hmaon@bumba.net.");
			fclose(f);
			fail = -1;
			break;
		}

		i >>= 1; /* i is number of bytes */

		if (!fd->h)
		{
			fd->dw = fd->w = 8; /* a nasty guess. */
			fd->dh = fd->h = i;
			/*if (i & 1) break;*/ /* odd number of bytes. this is the height. */
		} else 
		{
			if (i > fd->h) {
				fd->dw = fd->w = 16; /* an even nastier guess (full-width glyphs here) */
				if(i != (fd->h)*2) /* check sanity and file integrity */
				{
					plog("Error 2 in HEX measurement.");
					/*fail = -1;*/
				}
				break; /* this is a full-width glyph. We had the height. */
			} else
			if (i < fd->h) {
				if (i*2 != fd->h)
				{
					plog("Error 3 in HEX measurement.");
					/*fail = -1;*/
				}
				fd->dw = fd->w = 16; /* the same nastier guess. */
				fd->dh = fd->h = i; /* Ah, so this is the height */
			}
			/* they're equal. we can say nothing about the glyph height */
		}
	}

	/* analyze the file name */
	if(!strmetrics(filename, &mw, &mh))
	{
		/* success! */
		fd->dw = mw;
		fd->dh = mh;
	} else
	{
		plog("You may wish to incude the dimensions of a font in its file name. ie \"vga8x16.hex\"");
	}

	if (justmetrics) 
	{
		fclose(f);
		return fail;
	}

	/* Might as well allocate the bitmap here. */
	/* XXX I'm not sure what pixel format to use but the text blitter is 
	 * probably the wrong thing to start optimizing.
	 */
	fd->face = SDL_CreateRGBSurface(SDL_SWSURFACE, 256*fd->w, fd->h,8,0,0,0,0); 

	rewind(f);

	while (fgets(gs, MAX_HEX_FONT_LINE, f) != NULL)
	{
#ifdef FONT_LOAD_DEBUGGING
		puts(gs);
#endif
		/* figure out character code (aka index). xxxx:... */
		if (sscanf(gs, "%4x", &gx) != 1)
		{
			plog("Broken HEX file.");
			fail = -1;
			break;
		}

#ifdef FONT_LOAD_DEBUGGING
		printf("%4x:\n", gx);
#endif
		if (gx > 255) {
			gx = 255;
		}

		x = fd->w * gx; /* i is no longer necessary */
		y = 0;
		
		i = strlen(gs);

		if (gs[i-1] == '\n') {
			i--;
			gs[i] = '\0';
		}
		if (gs[i-1] == '\r') 
		{
			i--;
			gs[i] = '\0';
		}
		
		i -= 5; /* each line begins with 1234: */
		/* now i is the number of nibbles represented in the line */
		i >>= 1; /* now bytes. */

		pos = 5;
		bytesdone = 0; 

		while (gs[pos] != '\0' && pos < strlen(gs)) 
		{
			n  = (hexchartoi(gs[pos])) << 4; pos++; 
			n += (hexchartoi(gs[pos])); pos++;
			/* they're macros. do NOT pass a "pos++" to them! :) :) :) */

			for(j = 0; j < 8; ++j, ++x, n <<= 1)
			{
				if (n & 0x80) 
				{
#ifdef FONT_LOAD_DEBUGGING
					printf("#");
#endif
					((Uint8 *)fd->face->pixels)[x + y*fd->face->pitch] = 0xff;
				} else
				{
#ifdef FONT_LOAD_DEBUGGING
					printf("-");
#endif
					((Uint8 *)fd->face->pixels)[x + y*fd->face->pitch] = 0x00;
				}
			}
			++bytesdone;

			/* processing half-width glyph or just finished even byte */
			if (i == fd->h || ((i == 2*fd->h) && !(bytesdone & 1))) 
			{
				x = fd->w * gx;
				++y;
#ifdef FONT_LOAD_DEBUGGING
				printf("\n");
#endif
				if (y > fd->h) plog(format("y out of bounds. %d, %d", y, fd->h));
			} else if (i == 2*fd->h)
			{
				/* XXX do nothing? */
			} else 
			{
				/* XXX XXX XXX softer errors since HEX seems to actually permit
				 * this situation
				 */
				/*plog("HEX measurement error, fd->h is not a multiple of i.");*/
				/*fail = -1;*/
			}
		} /* while (gs[pos... */
	} /* while (fgets... */

	return fail;
}

		


#endif /* USE_HEX_FONTS */





/*** Function hooks needed by "Term" ***/

char *formatsdlflags(Uint32 flags) {
	return format ("%s%s%s%s%s%s%s%s%s%s (%x)", 
			(flags & SDL_HWSURFACE) ? "SDL_HWSURFACE "  : "",
			(flags & SDL_ANYFORMAT) ? "SDL_ANYFORMAT "  : "",
			(flags & SDL_HWPALETTE) ? "SDL_HWPALETTE "  : "",
			(flags & SDL_DOUBLEBUF) ? "SDL_DOUBLEBUF "  : "", 
			(flags & SDL_FULLSCREEN) ?"SDL_FULLSCREEN " : "",
			(flags & SDL_RESIZABLE) ? "SDL_RESIZABLE "  : "",
			(flags & SDL_HWACCEL) ?   "SDL_HWACCEL "    : "",
			(flags & SDL_SRCCOLORKEY) ? "SDL_SRCCOLRKEY "  : "",
			(flags & SDL_RLEACCEL) ? "SDL_RLEACCEL "   : "",
			(flags & SDL_SRCALPHA) ? "SDL_SRCALPHA "   : "",
			flags);
};


/*
 * XXX XXX XXX Init a new "term"
 *
 * This function should do whatever is necessary to prepare a new "term"
 * for use by the "term.c" package.  This may include clearing the window,
 * preparing the cursor, setting the font/colors, etc.  Usually, this
 * function does nothing, and the "init_xxx()" function does it all.
 */
static void Term_init_sdl(term *t)
{
	char asked[200];
	char got[200];
	term_data *td = (term_data*)(t->data);

	if (td->width && td->height) {
		/* TODO allow full-screen mode through command line */
		td->face = SDL_SetVideoMode(td->width, td->height, td->bpp, td->flags);
		if (td->face == NULL) {
			plog("SDL could not initialize video mode."); 
		}
		if (td->face->flags != td->flags) {
			strncpy(asked, formatsdlflags(td->flags), 199);
			asked[199] = 0;
			strncpy(got, formatsdlflags(td->face->flags), 199);
			got[199] = 0;
			plog(format("Vid. init.: We asked for %s and got %s", asked, got));
		}

	}

}



/*
 * XXX XXX XXX Nuke an old "term"
 *
 * This function is called when an old "term" is no longer needed.  It should
 * do whatever is needed to clean up before the program exits, such as wiping
 * the screen, restoring the cursor, fixing the font, etc.  Often this function
 * does nothing and lets the operating system clean up when the program quits.
 */
static void Term_nuke_sdl(term *t)
{
	term_data *td = (term_data*)(t->data);

	if (td->face) 
	{
		SDL_FreeSurface(td->face); /* what happen! someone set up us the bomb! */
		td->face = NULL;
	}
}



/*
 * XXX XXX XXX Do a "user action" on the current "term"
 *
 * This function allows the visual module to do things.
 *
 * This function is currently unused, but has access to the "info"
 * field of the "term" to hold an extra argument.
 *
 * In general, this function should return zero if the action is successfully
 * handled, and non-zero if the action is unknown or incorrectly handled.
 */
static errr Term_user_sdl(int n)
{
	term_data *td = (term_data*)(Term->data);

	/* XXX XXX XXX Handle the request */

	/* TODO What? Huh? */

	/* Unknown */
	return (1);
}



typedef struct sdl_keymapt sdl_keymapt;

struct sdl_keymapt {
	SDLKey k; /* what we get from SDL */
	char *s; /* what we feed to the Term_keypress */
	char *ctrl; /* what if CTRL is pressed? (NULL if the same) */
	char *shift; /* what if SHIFT is pressed? */
};

/* XXX XXX XXX the following keymap sucks. More comments below. */
sdl_keymapt sdl_keymap[] =
{
	{SDLK_UP, "[A", "Oa", "Ox"}, 
	{SDLK_DOWN, "[B", "Ob", "Or"},
	{SDLK_RIGHT, "[C", "Oc", "Ot"},
	{SDLK_LEFT, "[D", "Od", "Ov"},
	{SDLK_INSERT, "[2~", "[2^", "Op"},
	{SDLK_HOME, "[1~", "[1^", "Ow"},
	{SDLK_END, "[4~", "[4^", "Oq"},
	{SDLK_PAGEUP, "[5~", "[5^", "Oy"},
	{SDLK_PAGEDOWN, "[6~", "[6^", "Os"},
	{SDLK_F1, "[[A", NULL, NULL},
	{SDLK_F2, "[[B", NULL, NULL},
	{SDLK_F3, "[[C", NULL, NULL},
	{SDLK_F4, "[[D", NULL, NULL},
	{SDLK_F5, "[[E", NULL, NULL},
	{SDLK_F6, "[[17~", NULL, NULL},
	{SDLK_F7, "[[18~", NULL, NULL},
	{SDLK_F8, "[[19~", NULL, NULL},
	{SDLK_F9, "[[20~", NULL, NULL},
	{SDLK_F10, "[[21~", NULL, NULL},
	{SDLK_F11, "[[23~", NULL, NULL},
	{SDLK_F12, "[[24~", NULL, NULL},
	/* I have no machines with F13, F14, F15. Is that a Sun thing? */
	{SDLK_F13, "", NULL, NULL}, 
	{SDLK_F14, "", NULL, NULL},
	{SDLK_F15, "", NULL, NULL},
	{SDLK_RSHIFT, "", NULL, NULL}, 
	{SDLK_LSHIFT, "", NULL, NULL},
	{SDLK_RALT, "", NULL, NULL},
	{SDLK_LALT, "", NULL, NULL},
	{SDLK_RCTRL, "", NULL, NULL},
	{SDLK_LCTRL, "", NULL, NULL},
	{SDLK_RMETA, "", NULL, NULL},
	{SDLK_LMETA, "", NULL, NULL},
	{SDLK_NUMLOCK, "", NULL, NULL},
	{SDLK_CAPSLOCK, "", NULL, NULL},
	{SDLK_SCROLLOCK, "", NULL, NULL},
	{SDLK_LSUPER, "", NULL, NULL},
	{SDLK_RSUPER, "", NULL, NULL},
	{SDLK_HELP, "?", NULL, NULL},
	{SDLK_PRINT, "", NULL, NULL},
	{SDLK_SYSREQ, "", NULL, NULL},
	{SDLK_BREAK, "", NULL, NULL},
	{SDLK_MENU, "", NULL, NULL},
	{SDLK_POWER, "", NULL, NULL},
	{SDLK_EURO, "", NULL, NULL},
	{SDLK_0, "0", NULL, ")"}, /* XXX XXX XXX The CTRL-number keys need to be */
	{SDLK_1, "1", NULL, "!"}, /* defined since they represent digging for    */
	{SDLK_2, "2", NULL, "@"}, /* some people!. Really, this whole table      */
	{SDLK_3, "3", NULL, "#"}, /* should be replaced with something cleaner   */
	{SDLK_4, "4", NULL, "$"}, /* and an SDL pref file should be created.     */
	{SDLK_5, "5", NULL, "%"},
	{SDLK_6, "6", NULL, "^"},
	{SDLK_7, "7", NULL, "&"},
	{SDLK_8, "8", NULL, "*"},
	{SDLK_9, "9", NULL, "("},
	{SDLK_SEMICOLON, ";", NULL, ":"},
	{SDLK_COMMA, ",", NULL, "<"},
	{SDLK_PERIOD, ".", NULL, ">"},
	{SDLK_BACKSLASH, "\\", NULL, "|"},
	{SDLK_BACKQUOTE, "`", NULL, "~"},
	{SDLK_LEFTBRACKET, "[", NULL, "{"},
	{SDLK_RIGHTBRACKET, "]", NULL, "}"},
	{SDLK_MINUS, "-", NULL, "_"},
	{SDLK_EQUALS, "=", NULL, "+"},
	{SDLK_SLASH, "/", NULL, "?"},
	{SDLK_UNKNOWN, NULL, NULL, NULL} /* terminator */
};

void Multikeypress(char *k)
{
	while (*k) Term_keypress(*k++);
}


/*
 * XXX XXX XXX Do a "special thing" to the current "term"
 *
 * This function must react to a large number of possible arguments, each
 * corresponding to a different "action request" by the "term.c" package.
 *
 * The "action type" is specified by the first argument, which must be a
 * constant of the form "TERM_XTRA_*" as given in "term.h", and the second
 * argument specifies the "information" for that argument, if any, and will
 * vary according to the first argument.
 *
 * In general, this function should return zero if the action is successfully
 * handled, and non-zero if the action is unknown or incorrectly handled.
 */
static errr Term_xtra_sdl(int n, int v)
{
	term_data *td = (term_data*)(Term->data);

	SDL_Event event; /* just a temporary place to hold an event */

	/* Analyze */
	switch (n)
	{
		case TERM_XTRA_EVENT:

		/* XXX XXX XXX Process some pending events */
		/* Wait for at least one event if "v" is non-zero */
		/* otherwise, if no events are ready, return at once. */
		/* When "keypress" events are encountered, the "ascii" */
		/* value corresponding to the key should be sent to the */
		/* "Term_keypress()" function.  Certain "bizarre" keys, */
		/* such as function keys or arrow keys, may send special */
		/* sequences of characters, such as control-underscore, */
		/* plus letters corresponding to modifier keys, plus an */
		/* underscore, plus carriage return, which can be used by */
		/* the main program for "macro" triggers.  This action */
		/* should handle as many events as is efficiently possible */
		/* but is only required to handle a single event, and then */
		/* only if one is ready or "v" is true */
		/* This action is required. */

		do {
			if (v) 
			{
				if (!SDL_WaitEvent(&event)) return(0); /* TODO handle errors */
				v = 0;
			} else 
			{
				if(!SDL_PollEvent(&event)) return(0);
			}

			if (event.type == SDL_QUIT) 
			{
				/**((char *)0) = 1;*/ /* need crash data. */
				quit("Goodbye.");
			} else
			if (event.type == SDL_KEYDOWN) 
			{
				if (event.key.state == SDL_PRESSED) 
				{
					/* TODO If we use keysym.sym, we must make the program aware of
					 * the special SDLK_* keysym constants. Perhaps it can be done
					 * in pref files?
					 */
					/* For now, let's just simulate Linux getchar() */
					Uint32 i;
					Uint32 ch;

#ifdef SDL_MAGIC_KEYS
					/* Various ugly hacks. */
					switch(event.key.keysym.sym)
					{
						char buf[1024];
						FILE *tmp;
					/* Try to toggle graphics. */
					case SDLK_SCROLLOCK:
						use_graphics = !use_graphics;
						screen.t.higher_pict = !screen.t.higher_pict;
#ifdef reset_visuals
						reset_visuals();
#else
#warning Visuals cannot be reset.
#endif

#ifdef OLD_TERM_28X
						sprintf(buf, "%s-%s.prf", (use_graphics ? "graf" : "font"), ANGBAND_SYS);
#endif
						break;

					/* Try to save a screenshot. */
					case SDLK_PRINT:
						if (SDL_SaveBMP(screen.face, "newshot.bmp")) 
						{
							plog("You fail to get the screenshot off!");
							break;
						}
						for (i = 0; i < 999; ++i) {
							snprintf(buf, 1024, "%03d.bmp", i);
							if ((tmp = fopen(buf, "rb")) != NULL)
							{
								fclose(tmp);
								continue;
							}
							rename("newshot.bmp", buf);
						}
						plog("*click*");
						break;
					default:
						break;
					} /* switch */
#endif

					for (i = 0; ; ++i)
					{
						if (sdl_keymap[i].k == event.key.keysym.sym) 
						{
							if (sdl_keymap[i].s && strlen(sdl_keymap[i].s)) 
							{
								if (event.key.keysym.mod & KMOD_ALT) 
								{
									Term_keypress('');
								}
								if (event.key.keysym.mod & KMOD_CTRL)
								{
									if(sdl_keymap[i].ctrl) 
									{
										Multikeypress(sdl_keymap[i].ctrl);
										break;
									} 
								} else
								if (event.key.keysym.mod & KMOD_SHIFT)
								{
									if(sdl_keymap[i].shift)
									{
										Multikeypress(sdl_keymap[i].shift);
										break; 
									} 
								}
								Multikeypress(sdl_keymap[i].s);
							}
							break; /* out of the for() loop */
						} else
						if (sdl_keymap[i].k == SDLK_UNKNOWN)
						{
							if (event.key.keysym.mod & KMOD_ALT) 
							{
								Term_keypress('');
							}
							ch = event.key.keysym.sym; 
							/* alphanumeric keys aren't part of the keymap because
							 * typing them in would be way too tedious */
							if (ch <= 'z' && ch >= 'a') {
								if (event.key.keysym.mod & KMOD_CTRL)
								{
									ch = 1 + ch - 'a';
								} else
								if (event.key.keysym.mod & KMOD_SHIFT)
								{
									ch += ('A' - 'a');
								}
							}
							Term_keypress(ch);
							break; /* end the for loop; we're at the end of keymap */
						}
					} /* for... */
				} /* SDL_PRESSED */ 
			} 

		} while (SDL_PollEvent(NULL));

		return (0);

		case TERM_XTRA_FLUSH:

		/* XXX XXX XXX Flush all pending events */
		/* This action should handle all events waiting on the */
		/* queue, optionally discarding all "keypress" events, */
		/* since they will be discarded anyway in "term.c". */
		/* This action is required, but is often not "essential". */

		/*plog("TERM_XTRA_FLUSH");*/

		while (SDL_PollEvent(&event)) /* do nothing */ ;

		return (0);

		case TERM_XTRA_CLEAR:

		/* XXX XXX XXX Clear the entire window */
		/* This action should clear the entire window, and redraw */
		/* any "borders" or other "graphic" aspects of the window. */
		/* This action is required. */

		if (!td->face) return(1);

		/* a NULL dstrect fills the entire window */
		/* XXX the color 0 will not necessarily work for 8-bit modes */
		SDL_FillRect(td->face, NULL, 0);

		return (0);

		case TERM_XTRA_SHAPE:

		/* XXX XXX XXX Set the cursor visibility (optional) */
		/* This action should change the visibility of the cursor, */
		/* if possible, to the requested value (0=off, 1=on) */
		/* This action is optional, but can improve both the */
		/* efficiency (and attractiveness) of the program. */
		
		/* TODO implement me. */

		return (0);

		case TERM_XTRA_FROSH:

		/* XXX XXX XXX Flush a row of output (optional) */
		/* This action should make sure that row "v" of the "output" */
		/* to the window will actually appear on the window. */
		/* This action is optional on most systems. */

		if (!td->face) return -1;
		SDL_UpdateRect(td->face, 0, v*td->fd->dh, td->face->w, td->fd->dh);

		return (0);

		case TERM_XTRA_FRESH:

		/* XXX XXX XXX Flush output (optional) */
		/* This action should make sure that all "output" to the */
		/* window will actually appear on the window. */
		/* This action is optional if all "output" will eventually */
		/* show up on its own, or when actually requested. */

		if (!td->face) return -1;
		if (td->face->flags & SDL_HWSURFACE) SDL_Flip(td->face);

		return (0);

		case TERM_XTRA_NOISE:

		/* XXX XXX XXX Make a noise (optional) */
		/* This action should produce a "beep" noise. */
		/* This action is optional, but nice. */

		return (0);

		case TERM_XTRA_SOUND:

		/* XXX XXX XXX Make a sound (optional) */
		/* This action should produce sound number "v", where */
		/* the "name" of that sound is "sound_names[v]". */
		/* This action is optional, and not important. */

#ifdef USE_SDL_MIXER
		/* TODO make some noise. */
#else
		/* TODO We can actually make noise without the mixer too... */
#endif

		return (0);

		case TERM_XTRA_BORED:

		/* XXX XXX XXX Handle random events when bored (optional) */
		/* This action is optional, and not important */

		/* TODO add nifty effects here, I guess? */


		return (0);

		case TERM_XTRA_REACT:

		/* XXX XXX XXX React to global changes (optional) */
		/* For example, this action can be used to react to */
		/* changes in the global "color_table[256][4]" array. */
		/* This action is optional, but can be very useful */

		/* TODO garr, mateys, what have we got here? Er, nothing? Oh... */

		return (0);

		case TERM_XTRA_ALIVE:

		/* XXX XXX XXX Change the "hard" level (optional) */
		/* This action is used if the program changes "aliveness" */
		/* by being either "suspended" (v=0) or "resumed" (v=1) */
		/* This action is optional, unless the computer uses the */
		/* same "physical screen" for multiple programs, in which */
		/* case this action should clean up to let other programs */
		/* use the screen, or resume from such a cleaned up state. */
		/* This action is currently only used on UNIX machines */

		/* TODO this should probably do something... no, nevermind. */

		return (0);

		case TERM_XTRA_LEVEL:

		/* XXX XXX XXX Change the "soft" level (optional) */
		/* This action is used when the term window changes "activation" */
		/* either by becoming "inactive" (v=0) or "active" (v=1) */
		/* This action is optional but can be used to do things like */
		/* activate the proper font / drawing mode for the newly active */
		/* term window.  This action should NOT change which window has */
		/* the "focus", which window is "raised", or anything like that. */
		/* This action is optional if all the other things which depend */
		/* on what term is active handle activation themself. */

		return (0);

		case TERM_XTRA_DELAY:

		/* XXX XXX XXX Delay for some milliseconds (optional) */
		/* This action is important for certain visual effects, and also */
		/* for the parsing of macro sequences on certain machines. */

		SDL_Delay(v);

		return (0);
	}

	/* Unknown or Unhandled action */
	return (1);
}


/*
 * XXX XXX XXX Erase some characters
 *
 * This function should erase "n" characters starting at (x,y).
 *
 * You may assume "valid" input if the window is properly sized.
 */
static errr Term_wipe_sdl(int x, int y, int n)
{
	SDL_Rect dr;
	term_data *td = (term_data*)(Term->data);

	/*printf("%d, %d, %d\n", x, y, n);*/

	if (!td->gt || !td->fd) return 0;

	if (td->gt->dw != td->fd->dw || td->gt->dh != td->fd->dh)
	{
		dr.w = n * td->w;
		dr.h = td->h;
		dr.x = td->w * x;
		dr.y = td->h * y;
		SDL_FillRect(td->face, &dr, 0);
	}


	/* Success */
	return (0);
}


/*
 * XXX XXX XXX Display the cursor
 *
 * This routine should display the cursor at the given location
 * (x,y) in some manner.  On some machines this involves actually
 * moving the physical cursor, on others it involves drawing a fake
 * cursor in some form of graphics mode.  Note the "soft_cursor"
 * flag which tells "term.c" to treat the "cursor" as a "visual"
 * thing and not as a "hardware" cursor.
 *
 * You may assume "valid" input if the window is properly sized.
 *
 * You may use the "Term_grab(x, y, &a, &c)" function, if needed,
 * to determine what attr/char should be "under" the new cursor,
 * for "inverting" purposes or whatever.
 */
static errr Term_curs_sdl(int x, int y)
{
	/*term_data *td = (term_data*)(Term->data);*/

	/* XXX XXX XXX Display a cursor (see above) */
	
	/* XXX TODO implement a cursor. */

	/* Success */
	return (0);
}

/* The following draws one character to the Term, using font_data. */
inline static errr Term_char_sdl (int x, int y, byte a, unsigned char c){
	SDL_Rect sr, dr;
	term_data *td = (term_data*)(Term->data);

	if (!td->face) return -1;
	
	Term_wipe_sdl(x, y, 1);

	a &= 0x7f;
	c &= 0x7f;

	sr.w = td->fd->dw;
	dr.w = td->w;
	sr.h = td->fd->dh;
	dr.h = td->h;

	sr.x = c * td->fd->w;
	sr.y = 0;

	dr.x = x * td->w;
	dr.y = y * td->h;
	
	/* XXX Force SDL, or whatever it wraps, to make the text the color we want
	 * by tweaking the palette. This really is slower than blits between
	 * surfaces with identical color formats but it's so easy and convenient!
	 * Anyway, it seems to be fast enough on my machine. 
	 */
	SDL_SetColors(td->fd->face, &(color_data_sdl[a&0xf]), 0xff, 1);

	if(SDL_MUSTLOCK(td->face)) SDL_LockSurface(td->face);
	SDL_BlitSurface(td->fd->face, &sr, td->face, &dr);

	/* The following is only to test the function! 
	 * This is not how it will be used in the future!
	 */
	/*SDL_StretchBlit(td->fd->face, &sr, td->face, &dr);*/
	if(SDL_MUSTLOCK(td->face)) SDL_UnlockSurface(td->face);

	/* Success */
	return (0);
}

/* This one draws a graphical tile. */
inline static errr Term_tile_sdl (int x, int y, Uint8 a, Uint8 c){
	SDL_Rect sr, dr;
	/*int n;*/
	term_data *td = (term_data*)(Term->data);

	if(!td->face) return -1;

	Term_wipe_sdl(x, y, 1);

	sr.w = td->gt->w;
	sr.h = td->gt->h;

	dr.w = td->w;
	dr.h = td->h;

	sr.x = (c & 0x7F) * td->gt->w;
	sr.y = (a & 0x7F) * td->gt->h;


	dr.x = x * td->w;
	dr.y = y * td->h;

	if (sr.x > td->gt->face->w)
	{
		SDL_FillRect(td->face, &dr, SDL_MapRGB(td->face->format, 255, 64, 64));
		plog(format("OOBound (%d, %d) (%d, %d bitmap)", c&0x7f, a&0x7f, sr.x, sr.y)); 
	} else
	{
		SDL_BlitSurface(td->gt->face, &sr, td->face, &dr);
	}

	/* Success */
	return (0);
}


/*
 * XXX XXX XXX Draw a "picture" on the screen
 *
 * This routine should display the given attr/char pair at the
 * given location (x,y).  This function is only used if one of
 * the flags "always_pict" or "higher_pict" is defined.
 *
 * You must be sure that the attr/char pair, when displayed, will
 * erase anything (including any visual cursor) that used to be
 * at the given location.  On many machines this is automatic, on
 * others, you must first call "Term_wipe_xxx(x, y, 1)".
 *
 * With the "higher_pict" flag, this function can be used to allow
 * the display of "pseudo-graphic" pictures, for example, by using
 * "(a&0x7F)" as a "row" and "(c&0x7F)" as a "column" to index into
 * a special auxiliary pixmap of special pictures.
 *
 * With the "always_pict" flag, this function can be used to force
 * every attr/char pair to be drawn one at a time, instead of trying
 * to "collect" the attr/char pairs into "strips" with similar "attr"
 * codes, which would be sent to "Term_text_xxx()".
 *
 * This is the implementation of the text and 2D tile graphics display.
 * TODO Implement and wrap other graphics engines. They should replace
 * Term_pict_sdl as the hook in term, probably.
 *
 */

/* Wow, these #ifdefs look ugly. XXX */

#ifdef OLD_TERM_28X
static errr Term_pict_sdl(int x, int y, byte a, char c)
#else
static errr Term_pict_sdl(int x, int y, int n,  byte *ap, char *cp)
#endif
{
	term_data *td = (term_data*)(Term->data);

#ifdef OLD_TERM_28X
	byte *ap = &a;
	char *cp = &c;
#else
	while(n--)
	{
#endif

#ifdef USE_TILE_GRAPHICS
			if (td->gt && td->gt->face) /* it never hurts (much) to check */
			{
				Term_tile_sdl(x, y, *ap, *cp); /* draw a graphical tile */
			} else
#endif
		Term_char_sdl(x, y, *ap, *cp); /* print some text */

#ifdef OLD_TERM_28X
		/* nothing */
#else
	}
#endif

	/* Success */
	return (0);
}


/*
 * XXX XXX XXX Display some text on the screen
 *
 * This function should actually display a string of characters
 * starting at the given location, using the given "attribute",
 * and using the given string of characters, which is terminated
 * with a nul character and which has exactly "n" characters.
 *
 * You may assume "valid" input if the window is properly sized.
 *
 * You must be sure that the string, when written, erases anything
 * (including any visual cursor) that used to be where the text is
 * drawn.  On many machines this happens automatically, on others,
 * you must first call "Term_wipe_xxx()" to clear the area.
 *
 * You may ignore the "color" parameter if you are only supporting
 * a monochrome environment, unless you have set the "draw_blanks"
 * flag, since this routine is normally never called to display
 * "black" (invisible) text, and all other colors should be drawn
 * in the "normal" color in a monochrome environment.
 *
 * Note that this function must correctly handle "black" text if
 * the "always_text" flag is set, if this flag is not set, all the
 * "black" text will be handled by the "Term_wipe_xxx()" hook.
 */
static errr Term_text_sdl(int x, int y, int n, byte a, cptr s)
{
	/*term_data *td = (term_data*)(Term->data);*/

	while(n > 0) {
		Term_char_sdl(x, y, a, *s);
		++x; --n; ++s;
	}

	/* Success, we hope. */
	return (0);
}




/*
 * XXX XXX XXX Instantiate a "term_data" structure
 *
 * This is one way to prepare the "term_data" structures and to
 * "link" the various informational pieces together.
 *
 * This function assumes that every window should be 80x24 in size
 * (the standard size) and should be able to queue 256 characters.
 * Technically, only the "main screen window" needs to queue any
 * characters, but this method is simple.
 *
 * Note that "activation" calls the "Term_init_xxx()" hook for
 * the "term" structure, if needed.
 */
static void term_data_link(term_data *td)
{
	term *t = &td->t;

	/* Initialize the term */
	term_init(t, 80, 24, 256);

	/* XXX XXX XXX Choose "soft" or "hard" cursor */
	/* A "soft" cursor must be explicitly "drawn" by the program */
	/* while a "hard" cursor has some "physical" existance and is */
	/* moved whenever text is drawn on the screen.  See "term.c". */
	/* t->soft_cursor = TRUE; */

	/* XXX XXX XXX Avoid the "corner" of the window */
	/* t->icky_corner = TRUE; */

	/* XXX XXX XXX Use "Term_pict()" for all data */
	/* See the "Term_pict_xxx()" function above. */
	/* t->always_pict = TRUE; */

	/* XXX XXX XXX Use "Term_pict()" for "special" data */
	/* See the "Term_pict_xxx()" function above. */
	/* t->higher_pict = TRUE;*/ /* XXX Should this be set or not? */

	/* XXX XXX XXX Use "Term_text()" for all data */
	/* See the "Term_text_xxx()" function above. */
	/* t->always_text = TRUE; */

	/* XXX XXX XXX Ignore the "TERM_XTRA_BORED" action */
	/* This may make things slightly more efficient. */
	t->never_bored = TRUE;

	/* XXX XXX XXX Ignore the "TERM_XTRA_FROSH" action */
	/* This may make things slightly more efficient. */
	/* t->never_frosh = TRUE; */

	/* Erase with "white space" */
	t->attr_blank = TERM_WHITE;
	t->char_blank = ' ';

	/* Prepare the init/nuke hooks */
	t->init_hook = Term_init_sdl;
	t->nuke_hook = Term_nuke_sdl;

	/* Prepare the template hooks */
	t->user_hook = Term_user_sdl;
	t->xtra_hook = Term_xtra_sdl;
	t->wipe_hook = Term_wipe_sdl;
	t->curs_hook = Term_curs_sdl;
	t->pict_hook = Term_pict_sdl;
	t->text_hook = Term_text_sdl;

	/* Remember where we came from */
	t->data = (vptr)(td);

	/* Activate it */
	Term_activate(t);
}



/* #ifdef USE_SDL XXX redundant */

/*
 * A "normal" system uses "main.c" for the "main()" function, and
 * simply adds a call to "init_xxx()" to that function, conditional
 * on some form of "USE_XXX" define.
 */

/*
 * XXX XXX XXX Initialization function
 */
errr init_sdl(int argc, char **argv)
{
	term_data *td;
	Uint32 initflags = SDL_INIT_VIDEO; /* What's the point, if not video? */
	char path[1024];
	char fontname[64];
	char tilebmpname[64];
	char *a;
	int fh = 0, fw = 0; /* term_data.w and .h overrides */
	Uint32 gh = 13, gw = 8;  /* dimensions of graphical tilse.  */

	int fullscreen = 0;

	int bpp = 0;


#ifdef OLD_TERM_28X
	use_graphics = 0;
#else
	use_graphics = (arg_graphics == TRUE);
	ANGBAND_GRAF = "new"; /* not necessarily right.. set again below. XXX */
	ANGBAND_SYS = "sdl"; /* reduntant, hopefully */
#endif


	/* set some defaults before checking command line */
#ifdef USE_HEX_FONTS
#ifdef OLD_TERM_28X
	strcpy(fontname, "vga8x16.hex");
#else
	strcpy(fontname, "font/vga8x16.hex"); /* slashes are system specific!! FIXME */
#endif
#endif

#ifdef USE_TILE_GRAPHICS
#ifdef OLD_TERM_28X
	strcpy(tilebmpname, "16x16.bmp");
#else
	strcpy(tilebmpname, "graf/16x16.bmp");
#endif
#endif


#ifndef NO_SDL_COMMANDLINE
	--argc, ++argv; /* skip program name */
#define getarg ((argc-- > 0) ? (*(argv++)) : " ")
#define ungetarg (++argc, --argv)
	while(argc > 0) { 
		a = getarg;

#ifdef USE_HEX_FONTS
		if (!strcmp(a, "--hexfont"))
		{
			strncpy(fontname, getarg, 63);
			fontname[63] = '\0';
			/*if(strchr(fontname, '/')) 
			{
				plog("Disallowed character(s) in font filename.");
				strcpy(fontname, "vga.hex");
			}*/
		} else
#endif

#ifdef USE_TILE_GRAPHICS
		if (!strcmp(a, "--bmp") || !strcmp(a, "--graf")) 
		{
			sscanf(getarg, "%63s", tilebmpname);
		} else
#endif

		if (!strcmp(a, "--settilesize") || !strcmp(a,"--fwh")) 
		{
			sscanf(getarg, "%d", &fw);
			sscanf(getarg, "%d", &fh);
		} else

		if (!strcmp(a, "--fullscreen"))
		{
			fullscreen = 1;
		} else

		if (!strcmp(a, "--gfx") || !strcmp(a, "-g")) {
			use_graphics = 1;
		} else

		if (!strcmp(a, "--bpp"))
		{
			sscanf(getarg, "%d", &bpp);
		}

	}
#endif
		


	/* XXX XXX XXX Initialize the system */

#ifdef USE_SDL_MIXER
	initflags |= SDL_INIT_AUDIO;
	/* TODO perhaps use SDL_InitSubSystem() instead. */
#endif

	/* I don't think you'd want the following: */
#ifdef SDL_NOPARACHUTE
	initflags |= SDL_INIT_NOPARACHUTE;
#endif

	initflags |= SDL_INIT_EVENTTHREAD;


	if (SDL_Init(initflags) != 0) {
		/* how do we report an error here? */
		/*plog("SDL_init() returned an error.");*/
		return -1;	 /* like this. */
	}

	/* TODO figure out if we want atexit() or not */
	atexit(SDL_Quit); 

	init_color_data_sdl();

#ifdef OLD_TERM_28X
	strcpy(path,"lib/"); /* XXX XXX XXX this is system dependent!!!! FIXME */
	init_file_paths(path);
#else
	/*init_stuff(); XXX actually called in main.c. oops. */
#endif


	/* TODO think of a clever use for virtual extra windows */
	/* Recall window */
	td = &recall;
	WIPE(td, term_data);
	td->name = "Recall";
	td->face = NULL;
	td->width = td->height = td->bpp = 0;
	td->fd = &screen_font;
	/* XXX XXX XXX Extra stuff */
	term_data_link(&recall);
#ifdef OLD_TERM_28X
	term_recall = &recall.t;
#endif

	/* Choice window */
	td = &choice;
	WIPE(td, term_data);
	td->name = "Choice";
	td->face = NULL;
	td->width = td->height = td->bpp = 0;
	td->fd = &screen_font;
	/* XXX XXX XXX Extra stuff */
	term_data_link(&choice);
#ifdef OLD_TERM_28X
	term_choice = &choice.t;
#endif
	/* Mirror window */
	td = &mirror;
	WIPE(td, term_data);
	td->name = "Mirror";
	td->face = NULL;
	td->width = td->height = td->bpp = 0;
	td->fd = &screen_font;
	/* XXX XXX XXX Extra stuff */
	term_data_link(&mirror);
#ifdef OLD_TERM_28X
	term_mirror = &mirror.t;
#endif


	/* Screen window */
	td = &screen;
	WIPE(td, term_data);
	td->name = "Angband";
	td->face = NULL;

	td->fd = &screen_font;
	td->gt = &screen_tiles;

	/* XXX tile sizes will have to be figured out before any of the windows
	 * above can be initialized so move that code if you use them
	 * for something.
	 */
	/* XXX Hack. get font metrics and tile size to calculate window size. */
#ifdef USE_HEX_FONTS
	screen_font.w = screen_font.h = 0; screen_font.face = NULL;
	load_HEX_font_sdl(&screen_font, fontname, TRUE);
	if(td->w < td->fd->dw) td->w = td->fd->dw; 
	if(td->h < td->fd->dh) td->h = td->fd->dh; 
#endif

#ifdef USE_TILE_GRAPHICS
	screen_tiles.w = screen_tiles.dw = gw;
	screen_tiles.h = screen_tiles.dh = gh;
	if (strmetrics(tilebmpname, &gw, &gh))
	{
		/* strmetrics() has failed. keep the (likely wrong) default */
		plog(format("strmetrics() failed for %s", tilebmpname));
	} else
	{
		/*plog(format("%s yielded %d, %d by strmetrics()", tilebmpname, gw, gh));*/
		screen_tiles.w = screen_tiles.dw = gw;
		screen_tiles.h = screen_tiles.dh = gh;
		/* This will enlarge the window tile size to fit all elements.
		 * Of course, this might not be the behaviour you want.
		 */
	}
	if(td->w < screen_tiles.w) td->w = screen_tiles.w;
	if(td->h < screen_tiles.h) td->h = screen_tiles.h;
#endif

	/* Allow a manual override. It's all about the manual override. */
	if(fw && fh)
	{
		td->w = fw;
		td->h = fh;
	}

	/* XXX possibly, these should be calculated in Term_init_sdl.c */
	td->width = 80 * td->w; 
	td->height = 24 * td->h; 
	td->bpp = bpp; /* 0 is for current display bpp */

	if (fullscreen) td->flags |= SDL_FULLSCREEN | SDL_HWSURFACE;


	/* XXX XXX XXX Extra stuff */
	term_data_link(&screen);
	term_screen = &screen.t;

	if (screen.face == NULL) {
		plog("Shutting down SDL due to error(s).");
		SDL_Quit();
		return -1;
	}

	SDL_WM_SetCaption(screen.name, "Angband (SDL)");
	SDL_EnableKeyRepeat(SDL_DEFAULT_REPEAT_DELAY, SDL_DEFAULT_REPEAT_INTERVAL);

#ifdef USE_HEX_FONTS
	memset(&screen_font, 0, sizeof(font_data));
	screen_font.face = NULL; /* to be pedantic */
#ifdef FONT_LOAD_DEBUGGING
	plog("load_HEX...");
#endif
	if(load_HEX_font_sdl(&screen_font, fontname, 0)) 
	{
		plog("load_HEX_font_sdl() failed...");
		return -1;
	}
#else
#error No output method defined for SDL!
#endif

#ifdef USE_TILE_GRAPHICS

	path_build(path, 1023, ANGBAND_DIR_XTRA, tilebmpname);

	if((screen_tiles.face = SDL_LoadBMP(path)) == NULL)
	{
		plog(format("Sorry, could not load %s", path));
	} else
	{
		td->t.higher_pict = use_graphics;
		/*td->t.always_pict = use_tiles;*/
	}
#endif

	

	return 0;
}


/* #endif */ /* XXX redundant */ /* USE_SDL */

#endif
